#!/usr/bin/python -u
#
# PyGLiiPong - Pong for WMD with the Wii Remote
#
# Copyright (C) robomancer <robomancer@gmail.com>
#               e-Hernick  <onolog@gmail.com>
#
# LiiPong is free software; you can redistribute it and/or modify it
# under the terms of version 2 of the GNU General Public License, as
# published by the Free Software Foundation.
# 
# LiiPong is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with LiiPong; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Massive thanks to robomancer and others for the "LiiPong"
# code, which this is heavily based on.

from wmd.common import *
from wmd.UI.PyGCommon import PyGCommon

import time
import random

import pygame

def bound(low, num, high):
    """Returns:
    low if num < low
    high if num > high
    num otherwise"""
    return min(max(low, num), high)


class Pong(object):
    WIDTH = 640
    HEIGHT = 480
    PADDLE_SPEED = 5000.0
    PADDLE_HEIGHT = 100
    PADDLE_WIDTH = 16
    BALL_RADIUS = 10
    MAX_VY = 800
    
    class Ball(object):
        def __init__(self, x, y):
            self.initialX = x
            self.initialY = y
            self.reset()

        def reset(self):
            self.x = self.initialX
            self.y = self.initialY
            self.vx = 600.0
            self.vy = random.uniform(-Pong.MAX_VY / 4.0, Pong.MAX_VY / 4.0)

        def update(self, timeElapsed, paddles):
            self.x += timeElapsed * self.vx
            self.y += timeElapsed * self.vy
            result = -1
            if self.x < 0:
                result = 1
            elif self.x > Pong.WIDTH:
                result = 0
            if self.y < 0:
                self.y = 0
                self.vy = -self.vy
            elif self.y > Pong.HEIGHT:
                self.y = Pong.HEIGHT
                self.vy = -self.vy
            for p in paddles:
                if (abs(self.x - p.x) < Pong.BALL_RADIUS + Pong.PADDLE_WIDTH / 2 and
                    abs(self.y - p.y) < Pong.BALL_RADIUS + Pong.PADDLE_HEIGHT / 2):
                    self.vx = -self.vx
                    self.x = p.x + cmp(self.vx, 0) * (Pong.BALL_RADIUS + Pong.PADDLE_WIDTH)
                    self.vy = (self.y - p.y) * Pong.MAX_VY / Pong.PADDLE_HEIGHT
            return result

        def draw(self, surface):
            pygame.draw.circle(surface, (255, 0, 255),
                               (int(self.x), int(self.y)), Pong.BALL_RADIUS)

    class Paddle(object):
        
        def __init__(self, x, y, color):
            self.initialX = x
            self.initialY = y
            self.color = color
            self.reset()

        def reset(self):
            self.x = self.initialX
            self.y = self.initialY
            self.lastSpeed = 0

        def update(self, timeElapsed, speed):
            self.y += timeElapsed * speed * Pong.PADDLE_SPEED
            self.y = bound(0, self.y, Pong.HEIGHT)

        def draw(self, surface):
            halfWidth = Pong.PADDLE_WIDTH / 2
            halfHeight = Pong.PADDLE_HEIGHT / 2
            pygame.draw.rect(surface, self.color,
                             (int(self.x) - halfWidth, int(self.y) - halfHeight,
                              Pong.PADDLE_WIDTH, Pong.PADDLE_HEIGHT))
            
    def __init__(self, wiimotes):
        self.wiimotes = wiimotes
        self.ball = Pong.Ball(Pong.WIDTH / 2, Pong.HEIGHT / 2)
        self.paddles = [Pong.Paddle(25, Pong.HEIGHT / 2, (255, 255, 0)),
                        Pong.Paddle(Pong.WIDTH - 25, Pong.HEIGHT / 2, (0, 255, 255))]
        self.lastUpdate = time.time()
        self.scores = [0, 0]
        self.paused = True
        self.lastButtonState = 0
        self.lastSpeed = 0.0

    def start(self):
        pygame.init()
        pygame.display.set_caption("LiiPong")
        videoFlags = pygame.DOUBLEBUF
        self._screen = pygame.display.set_mode((Pong.WIDTH, Pong.HEIGHT), videoFlags)
        self.surface = pygame.display.get_surface()
        self.lastUpdate = time.time()

    def update(self):
        now = time.time()
        timeElapsed = now - self.lastUpdate
        self.lastUpdate = now

        # check for un-pause from left player
        button_a = self.wiimotes[0].get_button(pygame.wiimote.WM_A)
        if button_a != self.lastButtonState:
            self.lastButtonState = button_a
            if button_a:
                self.paused = not self.paused

        if not self.paused:
            # update based on wiimote tilt
            for i, wiimote in enumerate(self.wiimotes):
                command = wiimote.get_force()[1]
                speed = (command - 138) / 255.0
                speed = self.paddles[i].lastSpeed * 0.8 + speed * 0.2
                self.paddles[i].lastSpeed = speed
                self.paddles[i].y = speed * 2000 + Pong.HEIGHT / 2
            # AI for single-player game
            if len(self.wiimotes) == 1:
                speed = bound(-1.0, (self.ball.y - self.paddles[1].y) / 1000.0, 1.0)
                self.paddles[1].update(timeElapsed, speed)
            # check for a goal
            result = self.ball.update(timeElapsed, self.paddles)
            if result != -1:
                self.paused = True
                self.ball.reset()
                for paddle in self.paddles:
                    paddle.reset()
                self.scores[result] += 1
                print "Goal!  Score is now %d to %d." % tuple(self.scores)

        self.surface.fill((0, 0, 0))
        for paddle in self.paddles:
            paddle.draw(self.surface)
        self.ball.draw(self.surface)
        pygame.display.flip()


class PyGLiPong(PyGCommon):
  def __init__( self, ev ):
    evHandlers = [ [ WM_ACC, self.on_ACC_Report ] ]
    PyGCommon.__init__( self, ev, evHandlers )

def main():
    wiimotes = WiiMoteWrapper()
    pong = Pong(wiimotes)

    print "Press A to un-pause; Home to quit"
    try:
        pong.start()
        while not wiimotes[0].get_button(pygame.wiimote.WM_HOME):
            update = False
            for wiimote in wiimotes:
                if wiimote.update():
                    update = True
                    cycles += 1
            if update:
                pong.update()
    except KeyboardInterrupt:
        pass

    end_time = time.time()
    print "%u cycles in %u seconds (%u c/s)" % (cycles, end_time-start_time, cycles/(end_time-start_time))

