# WMD 0.2 - Output System

# The Output System sends data to the WiiMote.
# It changes and manages the current operating mode of the WiiMote.
# It does a good job, but it's limited
# It can't work with that many modes, and even though the ReportParser
#   would be some work to get working with all modes, it's the Output System
#   that would require the most significant rewrite and improvement.


from wmd.Common import *

import time, random
from copy import deepcopy

# These are the Wiimote control codes
# We don't actually use all of them
FEATURE_DISABLE = 0x00
FEATURE_ENABLE = 0x04

IR_MODE_OFF = 0
IR_MODE_STD = 1
IR_MODE_EXP = 3
IR_MODE_FULL = 5

CMD_SET_REPORT = 0x52

RID_LEDS = 0x11
RID_MODE = 0x12
RID_IR_EN = 0x13
RID_SPK_EN = 0x14
RID_STATUS = 0x15
RID_WMEM = 0x16
RID_RMEM = 0x17
RID_SPK = 0x18
RID_SPK_MUTE = 0x19
RID_IR_EN2 = 0x1A

MODE_BASIC = 0x30
MODE_ACC = 0x31
MODE_ACC_IR = 0x33
MODE_ACC_IR_EXT = 0x37
MODE_FULL = 0x3e

# These are commands for the wiimote
# We shouldn't put them here like that
commandcodes = {
  "vibrate_on": 0x521305,
  "vibrate_off": 0x521304,
  "leds_off": 0x521100,
  # Well, actually, these two are almost fine, but they should be functions using constants
  "accept_attachment": 0x52120033,
  'request_0x20_report': 0x521504
}

def hex2s(h):
  return  ('%x' % h).decode("hex")

class WiimoteMode:
  def __init__( self, ev, backend ):
    self.backend = backend
    self.ir = Mode_IR( self )
    self.leds = LEDs( self, ev )
    self.vibe = Vibrator( self, ev )

    # Here we poll the WiiMote for status, in order to switch to the proper mode if
    # an external controller is plugged in
    self.send_command( hex2s( commandcodes[ 'request_0x20_report'] ) )

  # Not so bad, it's simple and clean
  # Basic packing of data for sending
  # Should have a way to listen for a reply, though
  # Maybe by setting a callback on fail? This should be done with a decorator if at all.
  def send( self, cmd, report, data ):
    c = chr(cmd) + chr(report)
    for d in data:
      c += chr(d)
    self.send_command( c )

  # This function is awful, it needs to be tested too.. but it works well if given a delay to do its job
  def setmode( self, mode, cont, rmbl=0 ):
    aux = 0
    rmbl = 0
    if rmbl:
      aux |= 0x01
    if cont:
      aux |= 0x04
    self.send(CMD_SET_REPORT,RID_MODE,[aux,mode])

  # This has never been touched, and lots of data flow through here
  # size here is redundant, since we can just use len(data) if we want.
  def senddata( self, data, offset, size): # see writing to data: [[#On-board Memory].
    of1 = offset >> 24 & 0xFF #extract offset bytes
    of2 = offset >> 16 & 0xFF
    of3 = offset >> 8 & 0xFF
    of4 = offset & 0xFF
    data2 = data + [0]*(16-len(data)) # append zeros to pad data if less than 16 bytes
    if len(data2) > 16:
      data2 = data2[:16] # crop data if we have too much
    # format is [OFFSET (BIGENDIAN),SIZE,DATA (16bytes)]
    self.send(CMD_SET_REPORT,RID_WMEM,[of1,of2,of3,of4,size]+data2)

  # Simple relay function, sends the data to the backend, currently PyBlueZ
  def send_command( self, command ):
    self.backend.send_command( command ) 

  # Well, the argument format is awful, but it reads RAM quite alright.
  def read_ram( self, addr_high, addr_low, read_len, a0=0, a1=0 ):
    req = [ a0, a1, addr_high, addr_low, 0, read_len ]
    self.send( 0x52, 0x17, req )

  # Yuk..
  def accept_attachment( self ):
    self.send_command( hex2s( commandcodes[ 'accept_attachment' ] ) ) 
    self.ir.callseq( 'IanExt' )

  # Yuk.. Well, we need to redo most of the init sequence on disconnect anyway, so the goal is noble
  # This very function should be requesting a mode switch, and it should be checking back on its work
  # Actually, it reveals the need for a third agent that calls the pure protocol implementation.
  # A new class is needer, a WiiMote Protocol Manager...
  def disconnect_attachment( self ):
    self.ir.callseq( 'Ian' )

# It's the only part of the output system that receives events from the UI
# It should work in the same way that rumble does
# With a decorator function in common?
# See notes for class Vibrator:

# ACTION: REFACTOR
# PRIORITY: LOW
class LEDs:
  prevstate = [0,0,0,0]
  state = [0,0,0,0]

  def __init__( self, wiimoteMode, ev ):
    self.ev = ev
    self.wiimoteMode = wiimoteMode

    self.ev.subscribe( SET_LED, self.ev_set_led )

  def ev_set_led( self, payload ):
    led = payload[0]
    action = payload[1]

    if action == 'toggle':
      self.toggle(led)
    elif action == 'on':
      self.on(led)
    elif action == 'off':
      self.off(led)

  def send_led_command( self ):
    command_start = 0x521100
    command_end = 0x00

    if self.wiimoteMode.vibe.state == "on":
      command_end += 1

    for i in range(3):
      if self.state[i]:
        command_end += (i+1) * 0x10

    command = hex(command_start + command_end)
    cmd = command[2:].decode("hex")

    self.wiimoteMode.send_command( cmd )

  def on( self, led ):
    if self.state[led] != 1:
      self.state[led] = 1
      self.refresh()

  def off( self, led ):
    if self.state[led] != 0:
      self.state[led] = 0
      self.refresh()

  def toggle( self, led ):
    if self.state[led] == 0:
      self.on( led )
    elif self.state[led] == 1:
      self.off( led )

  def refresh( self ):
    self.send_led_command()

# Once a touted feature, it no longer works very well. It hasn't been tested in a while.
# It would be ideal to implement this on the new output system.

# ACTION: TEST, REFACTOR
# PRIORITY: LOW
class Vibrator:
  """OH YEAH"""

  state = "off" 

  def __init__( self, wiimoteMode, ev ):
    self.wiimoteMode = wiimoteMode
    self.ev = ev

  def on( self ):
    if self.state != "on":
      self.cmd_on()
      self.state = "on"

  def off( self ):
      self.cmd_off()
      self.state = "off"

  def toggle( self ):
    if self.state == "on":
      self.off()
    else:
      self.on()

  def cmd_on( self ):
    self.wiimoteMode.send_command(hex2s(commandcodes['vibrate_on']))

  def cmd_off( self ):
    self.wiimoteMode.send_command(hex2s(commandcodes['vibrate_off']))



# This shouldn't be called Mode_IR. It's a mode switcher. It calls predefined
# mode-switching sequences. It's called bit-banging. It's not a great way
# to talk to the WiiMote. We could improve it by checking replies, and by
# separating each part of the sequence, so that complete sequences could be
# built to enable only the necessary features.

# It switches between the two primary operating modes for WMD, IR+ACC and IR+EXT+ACC
# depending if we have or not an extension controller plugged in.
# The timing could be made quicker.

# ACTION: TOTAL REFACTOR
# NOTE: Research required
# PRIORITY: LOW
class Mode_IR:
  def __init__( self, wiimoteMode, sequence="Ian", state="off" ):
    self.sequences = {
      'Ian': self.seq_ian,
      'IanExt': self.seq_ian_ext,
      'Cliff': self.seq_cliff,
      'Marcan': self.seq_marcan
    }
    
    self.send = wiimoteMode.send
    self.senddata = wiimoteMode.senddata
    self.setmode = wiimoteMode.setmode

    if self.sequences.has_key( sequence ):
      self.sequence = self.sequences[sequence]

    self.state = "off"
    if state == "on":
      self.state = "on"
      self.sequence()

  def callseq( self, sequence ):
    if self.sequences.has_key( sequence ):
      seq = self.sequences[ sequence ]
      seq()

  def on( self ):
    self.state = "on"
    self.sequence()

  def off( self ):
    self.state = "off"
 
  def seq_marcan( self ):
    # this seems to be the minimal code to get it to work
    self.setmode(MODE_ACC_IR,0)
    self.send(CMD_SET_REPORT,RID_IR_EN,[FEATURE_ENABLE])
    self.send(CMD_SET_REPORT,RID_IR_EN2,[FEATURE_ENABLE])
    self.senddata([8],0x04B00030,1) # enable IR data out
    self.senddata([0x90],0x04B00006,1) # sensitivity constants (guessed, Cliff seems to have more data, but this works for me)
    self.senddata([0xC0],0x04B00008,1)
    self.senddata([0x40],0x04B0001A,1)
    self.senddata([IR_MODE_EXP],0x04B00033,1) # enable IR output with specified data format

  def seq_cliff( self ):
    # this is Cliff's version pythonified, probably more accurate as far as sensitivity. Works pretty much the same for me.
    self.setmode(MODE_ACC_IR,0)
    self.send(CMD_SET_REPORT,RID_IR_EN,[FEATURE_ENABLE])
    self.send(CMD_SET_REPORT,RID_IR_EN2,[FEATURE_ENABLE])
    self.senddata([1],0x04B00030,1) # seems to enable the IR peripheral
    self.senddata([0x02, 0x00, 0x00, 0x71, 0x01, 0x00, 0xaa, 0x00, 0x64],0x04B00000,9)
    self.senddata([0x63, 0x03],0x04B0001A,2)
    # this seems incorrect - for FULL IR mode, we must use FULL wiimote mode (0x3e).
    # otherwise the data is probably garbled.
    self.senddata([IR_MODE_FULL],0x04B00033,1) 
    self.senddata([8],0x04B00030,1) # Enable data output. Can be specified first it seems, we don't really need to be in mode 1.

  def seq_ian( self ):
    for i in range(2):
      self.setmode(MODE_ACC_IR,0)
      time.sleep(0.005)
      self.send(CMD_SET_REPORT,RID_IR_EN,[FEATURE_ENABLE])
      time.sleep(0.003)
      self.send(CMD_SET_REPORT,RID_IR_EN2,[FEATURE_ENABLE])

      dataset = [
	[ 0x04B00030, 0x01 ],
	[ 0x04B00030, 0x08 ],
	[ 0x04B00006, 0x90 ],
	[ 0x04B00008, 0xC0 ],
	[ 0x04B0001A, 0x40 ],
	[ 0x04B00033, 0x33 ],
	[ 0x04B00030, 0x08 ]
      ]

      for d in dataset:
        time.sleep(0.001)
	self.senddata( [ d[1] ], d[0], 1 )

    time.sleep(0.05)

  def seq_ian_ext( self ):
    for i in range(2):
      self.setmode(MODE_ACC_IR_EXT,0)
      time.sleep(0.005)
      self.send(CMD_SET_REPORT,RID_IR_EN,[FEATURE_ENABLE])
      time.sleep(0.003)
      self.send(CMD_SET_REPORT,RID_IR_EN2,[FEATURE_ENABLE])

      dataset = [
	[ 0x04B00030, 0x01 ],
	[ 0x04B00030, 0x08 ],
	[ 0x04B00006, 0x90 ],
	[ 0x04B00008, 0xC0 ],
	[ 0x04B0001A, 0x40 ],
	[ 0x04B00033, 0x33 ],
	[ 0x04B00030, 0x08 ],
	[ 0x04A40040, 0x00 ] # enable extcont
      ]

      for d in dataset:
        time.sleep(0.001)
	self.senddata( [ d[1] ], d[0], 1 )

    time.sleep(0.05)
